# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.api
import ifcopenshell.geom
import ifcopenshell.util.selector
import ifcopenshell.util.shape
import ifcopenshell.util.representation
import multiprocessing
from logging import Logger
from typing import Union


class Patcher:
    def __init__(
        self,
        src: str,
        file: ifcopenshell.file,
        logger: Logger,
        query: str = "IfcBeam",
        force_faceted_brep: bool = False,
    ):
        """Convert element body representations to tessellations or faceted breps

        Some software have bugs that result in incorrect geometry being
        imported for more complex representation types, such as CSGs. This
        patch converts elements matching your query into a tessellation (or
        faceted brep for IFC2X3) which is generally supported by all
        implementations. Note that styles and shape aspects are not preserved.

        See example bug: https://github.com/Autodesk/revit-ifc/issues/707

        :param query: Query string to filter out elements to convert, defaults to "IfcBeam"
        :type query: str
        :param force_faceted_brep: Force using IfcFacetedBreps instead of IfcPolygonalFaceSets,
            defaults to `False`
        :type force_faceted_brep: bool

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "TessellateElements", "arguments": ["IfcBeam", False]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.query = query
        self.force_faceted_brep = force_faceted_brep

    def patch(self):
        context = ifcopenshell.util.representation.get_context(self.file, "Model", "Body", "MODEL_VIEW")
        if not context:
            return

        elements = ifcopenshell.util.selector.filter_elements(self.file, self.query)

        if not elements:
            return

        settings = ifcopenshell.geom.settings()
        settings.set("context-ids", [context.id()])

        # Iterator can handle only products, so we filter out everything else (IfcTypeProducts).
        products = []
        non_products = []
        for element in elements:
            if element.is_a("IfcProduct"):
                products.append(element)
            else:
                non_products.append(element)

        def store_geometry(
            element: ifcopenshell.entity_instance,
            shape: Union[ifcopenshell.geom.ShapeType, ifcopenshell.geom.ShapeElementType],
        ) -> None:
            geometry = getattr(shape, "geometry", shape)
            v = [[x.tolist() for x in ifcopenshell.util.shape.get_vertices(geometry)]]
            f = [ifcopenshell.util.shape.get_faces(geometry)]
            replacements[element] = (v, f)

        iterator = ifcopenshell.geom.iterator(settings, self.file, multiprocessing.cpu_count(), include=products)
        replacements = {}
        if iterator.initialize():
            for shape in iterator:
                element = self.file.by_guid(shape.guid)
                store_geometry(element, shape)

        for element in non_products:
            representation = ifcopenshell.util.representation.get_representation(element, context)
            if not representation:
                continue
            shape = ifcopenshell.geom.create_shape(settings, representation)
            store_geometry(element, shape)

        # Do the replacements outside the iterator to prevent messing up iterator state.
        for element, geometry in replacements.items():
            v, f = geometry
            mesh = ifcopenshell.api.run(
                "geometry.add_mesh_representation",
                self.file,
                context=context,
                vertices=v,
                faces=f,
                force_faceted_brep=self.force_faceted_brep,
            )
            if element.is_a("IfcProduct"):
                representations = element.Representation.Representations
                representations = [r for r in representations if r.ContextOfItems != context]
                representations.append(mesh)
                element.Representation.Representations = representations
            else:  # IfcTypeProduct.
                rep_maps = element.RepresentationMaps
                for rep_map in rep_maps:
                    mapped_rep = rep_map.MappedRepresentation
                    if mapped_rep.ContextOfItems == context:
                        rep_map.MappedRepresentation = mesh



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


class Patcher:
    def __init__(self, src, file, logger, elevation=0):
        """Sets the reference elevation of all IfcSites

        To completely reference model coordinates, a reference elevation should
        be specified on the IfcSite. This is often omitted or not possible by
        proprietary BIM applications. This patch lets you set the reference
        elevation attribute explicitly.

        Note that this does not physically shift the Z coordinates of anything.
        The reference elevation is simply a numerical attribute.

        :param elevation: The elevation to set.
        :type elevation: float

        Example:

        .. code:: python

            # All IfcSites will have their reference elevation set to 42.
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "SetRefElevation", "arguments": [42]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.elevation = elevation

    def patch(self):
        project = self.file.by_type("IfcProject")[0]
        sites = self.find_decomposed_ifc_class(project, "IfcSite")
        for site in sites:
            site.RefElevation = float(self.elevation)

    def find_decomposed_ifc_class(self, element, ifc_class):
        results = []
        rel_aggregates = element.IsDecomposedBy
        if not rel_aggregates:
            return results
        for rel_aggregate in rel_aggregates:
            for part in rel_aggregate.RelatedObjects:
                if part.is_a(ifc_class):
                    results.append(part)
                results.extend(self.find_decomposed_ifc_class(part, ifc_class))
        return results



# IfcOpenShell - IFC toolkit and geometry engine
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcOpenShell.
#
# IfcOpenShell is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcOpenShell is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcOpenShell.  If not, see <http://www.gnu.org/licenses/>.

import ifcpatch
import ifcopenshell
import ifcopenshell.api.georeference
import ifcopenshell.util.placement
import ifcopenshell.util.shape_builder
import test.bootstrap
import tempfile
import numpy as np
from pathlib import Path
from typing import Optional


class TestMergeProject(test.bootstrap.IFC4):
    def setup_project(self, ifc_file: Optional[ifcopenshell.file] = None):
        prefix = None if ifc_file else "MILLI"
        if ifc_file is None:
            ifc_file = ifcopenshell.file(schema=self.file.schema)

        ifcopenshell.api.run("root.create_entity", ifc_file, ifc_class="IfcProject")
        unit = ifcopenshell.api.run("unit.add_si_unit", ifc_file, unit_type="LENGTHUNIT", prefix=prefix)
        ifcopenshell.api.run("unit.assign_unit", ifc_file, units=[unit])
        model = ifcopenshell.api.context.add_context(ifc_file, "Model")
        ifcopenshell.api.context.add_context(ifc_file, "Model", "Body", "MODEL_VIEW", parent=model)

        matrix = np.eye(4)
        matrix[:, 3] = (1, 2, 3, 1)
        wall = ifcopenshell.api.run("root.create_entity", ifc_file, ifc_class="IfcWall")
        ifcopenshell.api.run("geometry.edit_object_placement", ifc_file, product=wall, matrix=matrix, is_si=True)
        return ifc_file

    def test_run(self):
        self.file = self.setup_project(self.file)
        second_file = self.setup_project()
        temp_path = Path(tempfile.gettempdir()) / "second.ifc"
        second_file.write(temp_path)
        output = ifcpatch.execute({"file": self.file, "recipe": "MergeProject", "arguments": [str(temp_path)]})

        assert self.file == output
        assert len(output.by_type("IfcWall")) == 2
        wall1, wall2 = output.by_type("IfcWall")

        # test that units are converted
        placement1 = ifcopenshell.util.placement.get_local_placement(wall1.ObjectPlacement)
        placement2 = ifcopenshell.util.placement.get_local_placement(wall2.ObjectPlacement)
        to_tuple = lambda arr: tuple(map(tuple, arr))
        matrix = np.eye(4)
        matrix[:, 3] = (1, 2, 3, 1)
        assert to_tuple(placement1) == to_tuple(placement2) == to_tuple(matrix)

    def test_reusing_geometric_contexts(self):
        self.file = self.setup_project(self.file)
        second_file = self.setup_project()
        output = ifcpatch.execute({"file": self.file, "recipe": "MergeProject", "arguments": [second_file]})
        assert len(output.by_type("IfcGeometricRepresentationContext")) == 2

    def test_using_the_georeferencing_of_the_original_project(self):
        if self.file.schema == "IFC2X3":
            return
        self.file = self.setup_project(self.file)
        second_file = self.setup_project()
        ifcopenshell.api.georeference.add_georeferencing(self.file)
        ifcopenshell.api.georeference.add_georeferencing(second_file)
        output = ifcpatch.execute({"file": self.file, "recipe": "MergeProject", "arguments": [second_file]})
        assert len(output.by_type("IfcProjectedCRS")) == 1
        assert len(output.by_type("IfcMapConversion")) == 1

    def test_shifting_the_other_project_to_match_the_original_project_origin(self):
        self.file = self.setup_project(self.file)
        second_file = self.setup_project()
        ifcopenshell.api.georeference.add_georeferencing(self.file)
        xaa1, xao1 = ifcopenshell.util.geolocation.angle2xaxis(10)
        ifcopenshell.api.georeference.edit_georeferencing(
            self.file,
            coordinate_operation={"Eastings": 10, "Northings": 20, "XAxisAbscissa": xaa1, "XAxisOrdinate": xao1},
            projected_crs={"Name": "EPSG:1234"},
        )
        ifcopenshell.api.georeference.add_georeferencing(second_file)
        xaa, xao = ifcopenshell.util.geolocation.angle2xaxis(30)
        ifcopenshell.api.georeference.edit_georeferencing(
            second_file,
            coordinate_operation={"Eastings": 30000, "Northings": 40000, "XAxisAbscissa": xaa, "XAxisOrdinate": xao},
            projected_crs={"Name": "EPSG:0"},
        )

        builder1 = ifcopenshell.util.shape_builder.ShapeBuilder(self.file)
        builder2 = ifcopenshell.util.shape_builder.ShapeBuilder(second_file)
        o1 = builder1.create_axis2_placement_3d()
        o2 = builder2.create_axis2_placement_3d()
        body1 = ifcopenshell.util.representation.get_context(self.file, "Model", "Body", "MODEL_VIEW")
        body2 = ifcopenshell.util.representation.get_context(second_file, "Model", "Body", "MODEL_VIEW")

        # Original file is in meters
        wall1 = self.file.by_type("IfcWall")[0]
        m1 = ifcopenshell.util.placement.get_local_placement(wall1.ObjectPlacement)
        assert np.allclose(m1[:, 3], (1, 2, 3, 1))
        global_m1 = ifcopenshell.util.geolocation.auto_local2global(self.file, m1, should_return_in_map_units=False)
        assert np.allclose(global_m1[:, 3], (11.332, 21.796, 3, 1))

        block = self.file.createIfcCsgSolid(self.file.createIfcBlock(o1, 2, 2, 2))
        rep = builder1.get_representation(context=body1, items=[block])
        ifcopenshell.api.geometry.assign_representation(self.file, product=wall1, representation=rep)
        shape = ifcopenshell.geom.create_shape(ifcopenshell.geom.settings(), wall1)
        verts = ifcopenshell.util.shape.get_shape_vertices(shape, shape.geometry)
        assert np.any(np.all(np.isclose(np.array((1., 2., 3.)), verts), axis=1))
        assert np.any(np.all(np.isclose(np.array((3., 4., 5.)), verts), axis=1))

        # Second file is in millimeters with a different false origin
        wall1 = second_file.by_type("IfcWall")[0]
        m1 = ifcopenshell.util.placement.get_local_placement(wall1.ObjectPlacement)
        assert np.allclose(m1[:, 3], (1000, 2000, 3000, 1))
        global_m1 = ifcopenshell.util.geolocation.auto_local2global(second_file, m1, should_return_in_map_units=False)
        assert np.allclose(global_m1[:, 3], (31866, 41232, 3000, 1))

        block = second_file.createIfcCsgSolid(second_file.createIfcBlock(o2, 2000, 2000, 2000))
        rep = builder2.get_representation(context=body2, items=[block])
        ifcopenshell.api.geometry.assign_representation(second_file, product=wall1, representation=rep)
        shape = ifcopenshell.geom.create_shape(ifcopenshell.geom.settings(), wall1)
        verts = ifcopenshell.util.shape.get_shape_vertices(shape, shape.geometry)
        assert np.any(np.all(np.isclose(np.array((1., 2., 3.)), verts), axis=1))
        assert np.any(np.all(np.isclose(np.array((3., 4., 5.)), verts), axis=1))

        output = ifcpatch.execute({"file": self.file, "recipe": "MergeProject", "arguments": [second_file]})

        # In the future we may use proj to support reprojection from different CRSes. For now... nope!
        if self.file.schema != "IFC2X3":
            assert output.by_type("IfcProjectedCRS")[0].Name == "EPSG:1234"

        # The results should be in meters with the false origin of the original file
        params = ifcopenshell.util.geolocation.get_helmert_transformation_parameters(output)
        assert params.e == 10
        assert params.n == 20
        assert params.xaa == xaa1
        assert params.xao == xao1

        wall1, wall2 = output.by_type("IfcWall")
        m1 = ifcopenshell.util.placement.get_local_placement(wall1.ObjectPlacement)
        m2 = ifcopenshell.util.placement.get_local_placement(wall2.ObjectPlacement)
        assert np.allclose(m1[:, 3], (1, 2, 3, 1))
        # assert np.allclose(m2[:, 3], (8.321, 29.321, 3, 1), atol=1e-3)
        assert np.allclose(m2[:, 3], (17.847, 24.707, 3., 1.), atol=1e-3)

        shape = ifcopenshell.geom.create_shape(ifcopenshell.geom.settings(), wall1)
        verts = ifcopenshell.util.shape.get_shape_vertices(shape, shape.geometry)
        assert np.any(np.all(np.isclose(np.array((1., 2., 3.)), verts), axis=1))
        assert np.any(np.all(np.isclose(np.array((3., 4., 5.)), verts), axis=1))

        shape = ifcopenshell.geom.create_shape(ifcopenshell.geom.settings(), wall2)
        verts = ifcopenshell.util.shape.get_shape_vertices(shape, shape.geometry)
        assert np.any(np.all(np.isclose(np.array((17.847, 24.707, 3.)), verts, atol=1e-3), axis=1))
        assert np.any(np.all(np.isclose(np.array((20.410, 25.902, 5.)), verts, atol=1e-3), axis=1))


class TestMergeProjectIFC2X3(test.bootstrap.IFC2X3, TestMergeProject):
    pass



# IfcOpenShell - IFC toolkit and geometry engine
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcOpenShell.
#
# IfcOpenShell is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcOpenShell is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcOpenShell.  If not, see <http://www.gnu.org/licenses/>.

import test.bootstrap
import ifcopenshell.api
import ifcopenshell.util.unit
import ifcpatch


class TestConvertLengthUnit(test.bootstrap.IFC4):
    # NOTE: conversion itself is covered by `ifcopenshell.util.unit.convert_file_length_units` tests
    def test_run(self):
        project = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcProject")
        unit = ifcopenshell.api.run("unit.add_si_unit", self.file, unit_type="LENGTHUNIT", prefix="MILLI")
        ifcopenshell.api.run("unit.assign_unit", self.file, units=[unit])
        output = ifcpatch.execute(
            {"input": "input.ifc", "file": self.file, "recipe": "ConvertLengthUnit", "arguments": ["METER"]}
        )
        unit = ifcopenshell.util.unit.get_project_unit(output, "LENGTHUNIT")
        assert ifcopenshell.util.unit.get_full_unit_name(unit) == "METRE"


class TestConvertLengthUnitIFC2X3(test.bootstrap.IFC2X3, TestConvertLengthUnit):
    pass



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import logging
import numpy as np
import numpy.typing as npt
import ifcopenshell
from typing import Literal, Optional, Union


class Patcher:
    def __init__(
        self,
        src: str,
        file: ifcopenshell.file,
        logger: logging.Logger,
        mode: Literal[
            "geometry",
            "placement",
            "both",
        ] = "geometry",
        a: Optional[float] = None,
        b: Optional[float] = None,
        c: Optional[float] = None,
        d: Optional[float] = None,
    ):
        """Reset any large coordinates to smaller coordinates based on a threshold

        If you find large coordinates in your model, the large coordinates may
        either by due to large coordinates in the object placement matrix of
        each object, or large coordinates in the geometry of each object.

        If it is the former (preferred), consult the OffsetObjectPlacements
        recipe. If it is the latter, this indicates seriously incorrect
        coordinates in your IFC model.

        This recipe finds any large coordinates (if either the X, Y, or Z
        ordinate is larger than a threshold) and offsets it back down to a small
        number.

        You may either manually specify the offset to apply to any large
        coordinate, or an offset will be automatically determined arbitrarily
        based on the first large number we encounter.

        Note that if your model inconsistently mixes coordinates between large
        and small (such as if your model mixes both local and map coordinates)
        then the results of this function may be poor. Provide a bug report back
        to your BIM application to get it fixed.

        You may specify up to 4 arguments, a, b, c, and d.

        If you specify no arguments, then the threshold is set to 1000000. The
        offset is auto detected.

        If you only specify 1 parameter (i.e. a), then this is treated as the
        threshold beyond which an ordinate is considered to be large. The offset
        is auto detected.

        If you specify 3 parameters, (i.e. a, b, c) then your three numbers are
        treated as the X, Y, Z offset to apply. Typically your numbers will be
        negative to bring the numbers smaller. The threshold is set to 1000000.

        If you specify 4 parameters (i.e. a, b, c, d), then the first three
        numbers are treated as the X, Y, Z offset to apply (a, b, c). The fourth
        (d) will be treated as the threshold.

        :param mode: Choose from "geometry", "placement", or "both". Choosing
            "geometry" will only replace cartesian points used in shape
            representations. Choosing "placement" will only replace cartesian
            points used in object placements. Choosing "both" will replace all
            cartesian points regardless of use (useful if the model has both
            large placement offsets and large geometry offsets).
        :type mode: str
        :param a: The first parameter
        :type a: float,optional
        :param b: The second parameter
        :type b: float,optional
        :param c: The third parameter
        :type c: float,optional
        :param d: The fourth parameter
        :type d: float,optional

        Example:

        .. code:: python

            # Reset all coordinates with an ordinate larger than 1000000 arbitrarily
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ResetAbsoluteCoordinates", "arguments": []})

            # Reset all coordinates with an ordinate larger than 1000 arbitrarily
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ResetAbsoluteCoordinates", "arguments": [1000]})

            # Reset all coordinates with an ordinate larger than 1000000 by -50000,-20000,0
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ResetAbsoluteCoordinates", "arguments": [-50000,-20000,0]})

            # Reset all coordinates with an ordinate larger than 1000 by -500,-200,0
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ResetAbsoluteCoordinates", "arguments": [-500,-200,0,1000]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.mode = mode
        self.args = [x for x in [a, b, c, d] if x is not None]

    def patch(self):
        placement_coord_ids = set()
        for placement in self.file.by_type("IfcObjectPlacement"):
            [placement_coord_ids.add(e.id()) for e in self.file.traverse(placement) if e.is_a("IfcCartesianPoint")]

        # Arbitrary threshold based on experience
        self.threshold = 1000000
        if self.args and len(self.args) == 1:
            self.threshold = float(self.args[0])
        elif self.args and len(self.args) == 4:
            self.threshold = float(self.args[3])

        # This method will not work all the time, but will catch most issues. It
        # assumes that absolute coordinates are easily recognisable based on
        # having a large absolute value above a threshold. This is not always
        # the case, but is very fast to run, and works for most cases.
        offset_point = None
        if self.args and len(self.args) >= 3:
            offset_point = (float(self.args[0]), float(self.args[1]), float(self.args[2]))
        try:
            point_lists = self.file.by_type("IfcCartesianPointList3D")
        except:
            # IFC2X3 does not have IfcCartesianPointList3D
            point_lists = []
        for point_list in point_lists:
            coord_list = [None] * len(point_list.CoordList)
            for i, point in enumerate(point_list.CoordList):
                if len(point) == 2 or not self.is_point_far_away(point):
                    coord_list[i] = point
                    continue
                if not offset_point:
                    offset_point = (-point[0], -point[1], -point[2])
                    self.logger.info(f"Resetting absolute coordinates by {point}")
                point = (point[0] + offset_point[0], point[1] + offset_point[1], point[2] + offset_point[2])
                coord_list[i] = point
            point_list.CoordList = coord_list
        for point in self.file.by_type("IfcCartesianPoint"):
            if len(point.Coordinates) == 2 or not self.is_point_far_away(point):
                continue
            if self.mode == "geometry":
                if point.id() in placement_coord_ids:
                    continue
            elif self.mode == "placement":
                if point.id() not in placement_coord_ids:
                    continue
            if not offset_point:
                offset_point = (-point.Coordinates[0], -point.Coordinates[1], -point.Coordinates[2])
                self.logger.info(f"Resetting absolute coordinates by {point}")
            point.Coordinates = (
                point.Coordinates[0] + offset_point[0],
                point.Coordinates[1] + offset_point[1],
                point.Coordinates[2] + offset_point[2],
            )

    def is_point_far_away(self, point: Union[ifcopenshell.entity_instance, npt.NDArray[np.float64]]) -> bool:
        if hasattr(point, "Coordinates"):
            return (
                abs(point.Coordinates[0]) > self.threshold
                or abs(point.Coordinates[1]) > self.threshold
                or abs(point.Coordinates[2]) > self.threshold
            )
        return abs(point[0]) > self.threshold or abs(point[1]) > self.threshold or abs(point[2]) > self.threshold



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy as np
import ifcopenshell
import ifcopenshell.util.placement


class Patcher:
    def __init__(self, src, file, logger, x=None, y=None, z=None, should_rotate_first=True, ax=None, ay=None, az=None):
        """Offset and rotate all object placements in a model

        Every physical object in an IFC model has an object placement, a
        matrix dictating where it is in XYZ space and its rotation.

        Sometimes, models will have their models offset incorrectly into map
        coordinates (i.e. very large coordinates) when they should be using
        local coordinates, or vice versa, or simply be using wrong coordinates.

        In some cases, models will even be rotated, especially with mixups where
        Y is up instead of Z, coming from low quality BIM software.

        This patch lets you translate, and optionally rotate (either rotate 2D
        in plan view along the Z axis, or rotate in 3D across any axis) the
        entire IFC model.

        :param x: The X coordinate to offset by in project length units.
        :type x: float
        :param y: The Y coordinate to offset by in project length units.
        :type y: float
        :param z: The Z coordinate to offset by in project length units.
        :type z: float
        :param should_rotate_first: Whether or not to rotate first and then
            translate, or to first translate and rotate afterwards. Defaults to
            rotate first then translate.
        :type should_rotate_first: bool
        :param ax: An optional angle to rotate by. If only this angle is
            specified, it is treated as the angle to rotate in plan view (i.e.
            around the Z axis). If all angle parameters are specified, then it
            is treated as the angle to rotate around the X axis. Angles are in
            decimal degrees and positive is anticlockwise.
        :type ax: float,optional
        :param ay: An optional angle to rotate by for 3D rotations along the Y
            axis. Angles are in decimal degrees and positive is anticlockwise.
        :type ay: float,optional
        :param az: An optional angle to rotate by for 3D rotations along the Z
            axis. Angles are in decimal degrees and positive is anticlockwise.
        :type az: float,optional

        Example:

        .. code:: python

            # Offset a model by 100 units in both the X and Y axis.
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "OffsetObjectPlacements", "arguments": [100,100,0]})

            # Rotate by 90 degrees, but don't do any offset
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "OffsetObjectPlacements", "arguments": [0,0,0,True,90]})

            # Some crazy 3D rotation and offset
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "OffsetObjectPlacements", "arguments": [12.5,5,2,False,90,90,45]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.x = x
        self.y = y
        self.z = z
        self.should_rotate_first = should_rotate_first
        self.ax = ax
        self.ay = ay
        self.az = az
        if self.ax is None:
            self.angle_type = None
        elif self.ay is None:
            self.angle_type = "2D"
        else:
            self.angle_type = "3D"

    def patch(self):
        absolute_placements = []

        for product in self.file.by_type("IfcProduct"):
            if not product.ObjectPlacement:
                continue
            absolute_placement = self.get_absolute_placement(product.ObjectPlacement)
            if absolute_placement.is_a("IfcLocalPlacement"):
                absolute_placements.append(absolute_placement)
        absolute_placements = set(absolute_placements)

        translate = self.identity_matrix()
        rotate = self.identity_matrix()

        if self.angle_type == "2D":
            angle = float(self.ax)
            if angle:
                rotate = self.z_rotation_matrix(math.radians(angle), rotate)
        elif self.angle_type == "3D":
            for arg in (("x", float(self.ax)), ("y", float(self.ay)), ("z", float(self.az))):
                if arg[1]:
                    rotate = getattr(self, f"{arg[0]}_rotation_matrix")(math.radians(arg[1]), rotate)

        translate[0][3] += float(self.x)
        translate[1][3] += float(self.y)
        translate[2][3] += float(self.z)

        if self.should_rotate_first:
            transformation = translate @ rotate
        else:
            transformation = rotate @ translate

        for placement in absolute_placements:
            placement.RelativePlacement = self.get_relative_placement(
                transformation @ ifcopenshell.util.placement.get_local_placement(placement)
            )

    def get_absolute_placement(self, object_placement):
        if object_placement.PlacementRelTo:
            return self.get_absolute_placement(object_placement.PlacementRelTo)
        return object_placement

    def identity_matrix(self):
        return np.eye(4)

    def x_rotation_matrix(self, angle, transformation):
        transformation[1][1] = math.cos(angle)
        transformation[1][2] = -math.sin(angle)
        transformation[2][1] = math.sin(angle)
        transformation[2][2] = math.cos(angle)
        return transformation

    def y_rotation_matrix(self, angle, transformation):
        transformation[0][0] = math.cos(angle)
        transformation[0][2] = math.sin(angle)
        transformation[2][0] = -math.sin(angle)
        transformation[2][2] = math.cos(angle)
        return transformation

    def z_rotation_matrix(self, angle, transformation):
        transformation[0][0] = math.cos(angle)
        transformation[0][1] = -math.sin(angle)
        transformation[1][0] = math.sin(angle)
        transformation[1][1] = math.cos(angle)
        return transformation

    def get_relative_placement(self, m):
        x = np.array((m[0][0], m[1][0], m[2][0]))
        z = np.array((m[0][2], m[1][2], m[2][2]))
        o = np.array((m[0][3], m[1][3], m[2][3]))
        object_matrix = ifcopenshell.util.placement.a2p(o, z, x)
        return self.create_ifc_axis_2_placement_3d(
            object_matrix[:, 3][0:3],
            object_matrix[:, 2][0:3],
            object_matrix[:, 0][0:3],
        )

    def create_ifc_axis_2_placement_3d(self, point, up, forward):
        return self.file.createIfcAxis2Placement3D(
            self.file.createIfcCartesianPoint(point.tolist()),
            self.file.createIfcDirection(up.tolist()),
            self.file.createIfcDirection(forward.tolist()),
        )



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


class Patcher:
    def __init__(self, src, file, logger, is_solid=True):
        """Fix missing or spot-coordinate bugged TINs loading in Revit

        TINs exported from 12D or Civil 3D may contain dense or highly obtuse
        triangles.  Although these will load in Revit, you will not be able to
        use Revit's Spot Coordinate or Spot Elevation tool.

        See bug: https://github.com/Autodesk/revit-ifc/issues/511

        If `is_solid` is enabled, we assume the surface is represented as a
        solid (e.g. I have come across surfaces which are extruded by 1mm from
        Civil 3D). The solution will delete any faces with a Z normal less
        than 0.5. In case the mesh has any side faces or thickness, this should
        leave only the top surface which is relevant for spot coordinates and
        elevations.

        If `is_solid` is disabled, then we assume the surface has no thickness.
        In this scenario, all faces with any normals pointing in a negative
        global Z direction will be flipped. This is because Revit cannot
        annotate backfaces.

        Vertices closer than 10mm will also be merged to prevent dense
        portions of the TIN at a minor sacrifice of surveying accuracy. It will
        also triangulate all meshes to prevent non-coplanar surfaces, and
        delete any obtuse triangles where one of their XY angles is less than
        0.3 degrees.  Therefore the result will contain some minor "holes" in
        the TIN, but these holes will only be in dense triangles that Revit
        can't handle anyway and won't affect most coordination tasks.

        Note that you may may want to run other tools like
        OffsetObjectPlacements or ResetAbsoluteCoordinates to fix large
        coordinates as these can also cause issues in Revit (such as inaccuracy
        or inability to use the Spot Coordinate / Elevation tool).

        This patch is designed to work on any TIN-like export, typically coming
        from civil software. It also requires you to run it using Blender, as
        the geometric modification uses the Blender geometry engine.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "FixRevitTINs", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.is_solid = is_solid

    def patch(self):
        import bpy
        import bmesh
        import blenderbim.tool as tool
        from math import degrees

        bpy.context.scene.BIMProjectProperties.should_use_native_meshes = True
        bpy.ops.bim.load_project(filepath=self.src)

        old_history_size = tool.Ifc.get().history_size
        old_undo_steps = bpy.context.preferences.edit.undo_steps
        tool.Ifc.get().history_size = 0
        bpy.context.preferences.edit.undo_steps = 0

        angle_threshold = 0.3

        for obj in bpy.data.objects:
            if not obj.BIMObjectProperties.ifc_definition_id or not obj.data:
                continue
            if not obj.data.polygons:
                continue
            data = obj.data
            bm = bmesh.new()
            bm.from_mesh(data)

            bm.faces.ensure_lookup_table()
            faces_to_delete = []

            if self.is_solid:
                for face in bm.faces:
                    if face.normal.z < 0.5:
                        faces_to_delete.append(face)

            bmesh.ops.delete(bm, geom=faces_to_delete, context='FACES_ONLY')

            bmesh.ops.remove_doubles(bm, verts=bm.verts, dist=0.01)
            bmesh.ops.triangulate(bm, faces=bm.faces[:], quad_method="BEAUTY", ngon_method="BEAUTY")
            bm.faces.ensure_lookup_table()
            for polygon in bm.faces:
                try:
                    v1, v2, v3 = [v.co.to_2d() for v in polygon.verts]
                    d1 = degrees((v2 - v1).angle(v3 - v1))
                    d2 = degrees((v3 - v2).angle(v1 - v2))
                    d3 = degrees((v1 - v3).angle(v2 - v3))
                    if d1 < angle_threshold or d2 < angle_threshold or d3 < angle_threshold:
                        bm.faces.remove(polygon)
                except:
                    bm.faces.remove(polygon)
            if bm.faces:
                if not self.is_solid:
                    bmesh.ops.recalc_face_normals(bm, faces=bm.faces)
                    for face in bm.faces:
                        global_normal = obj.matrix_world.to_3x3() @ face.normal
                        if global_normal.z < 0:
                            face.normal_flip()

                bm.to_mesh(data)
                bm.free()
                bpy.ops.bim.update_representation(obj=obj.name, ifc_representation_class="")

        tool.Ifc.get().history_size = old_history_size
        bpy.context.preferences.edit.undo_steps = old_undo_steps

        self.file = tool.Ifc.get()



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.api.georeference
from ifcpatch.recipes import OffsetObjectPlacements, SetWorldCoordinateSystem


class Patcher:
    def __init__(
        self,
        src,
        file,
        logger,
        name: str = "EPSG:1234",
        x: float = 0,
        y: float = 0,
        z: float = 0,
        e: float = 0,
        n: float = 0,
        h: float = 0,
        gn_angle: float = 0,
        rotate_angle: float = 0,
    ):
        """Sets a false origin with a map conversion in a model

        On IFC2X3 models, a EPset_MapConversion is used.

        :param x: The local X coordinate which will become the new false origin.
        :param y: The local Y coordinate which will become the new false origin.
        :param z: The local Z coordinate which will become the new false origin.
        :param e: The easting which the false origin correlates to.
        :param n: The northing which the false origin correlates to.
        :param h: The height which the false origin correlates to.
        :param gn_angle: The anticlockwise angle to grid north.
        :param rotate_angle: An anticlockwise angle to rotate the model by if
            necessary (pivoted by the false origin).

        Example:

        .. code:: python

            # Set the current origin 0,0,0 to correlate to map coordinates 1000,1000,0 and a grid north of 15.
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "SetFalseOrigin", "arguments": ["EPSG:1234", 0, 0, 0, 1000, 1000, 0, 15, 0]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.name = name
        self.x = x
        self.y = y
        self.z = z
        self.e = e
        self.n = n
        self.h = h
        self.gn_angle = gn_angle
        self.rotate_angle = rotate_angle

    def patch(self):
        SetWorldCoordinateSystem.Patcher(self.src, self.file, self.logger, x=0, y=0, z=0, ax=0, ay=0, az=0).patch()
        coordinate_operation = {
            "Eastings": self.e,
            "Northings": self.n,
            "OrthogonalHeight": self.h,
        }
        if self.gn_angle:
            a, o = ifcopenshell.util.geolocation.angle2xaxis(self.gn_angle)
            coordinate_operation.update({"XAxisAbscissa": a, "XAxisOrdinate": o})
        ifc_class = "IfcMapConversionScaled"
        if self.file.schema in ("IFC4", "IFC2X3"):
            ifc_class = "IfcMapConversion"
        ifcopenshell.api.georeference.add_georeferencing(self.file, ifc_class=ifc_class)
        ifcopenshell.api.georeference.edit_georeferencing(
            self.file, projected_crs={"Name": self.name}, coordinate_operation=coordinate_operation
        )
        OffsetObjectPlacements.Patcher(
            self.src,
            self.file,
            self.logger,
            x=-self.x,
            y=-self.y,
            z=self.z,
            should_rotate_first=False,
            ax=self.rotate_angle or None,
        ).patch()



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.util.element


class Patcher:
    def __init__(self, src, file, logger):
        """Downgrade indexed polycurves to simple polylines

        Low quality IFC viewers like Navisworks do not support various IFC4
        geometry, such as indexed polycurves. These can result in missing
        geometry or geometric glitches (such as arcs being displayed as full
        circles). This is pretty common when viewing IFCs from ArchiCAD that
        include site boundaries (incorrectly drawn using the ArchiCAD grid tool,
        as ArchiCAD has no site boundary tool).

        This will downgrade specifically the indexed polycurve geometry types in
        an IFC4 model (IFC2X3 does not have this geometry type) to help
        compatibility in viewers like Navisworks.

        Example:

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "DowngradeIndexedPolyCurve", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        if self.file.schema == "IFC2X3":
            return
        curve_map = {}

        for curve in self.file.by_type("IfcIndexedPolyCurve"):
            if "IfcArcIndex" in [s.is_a() for s in curve.Segments]:
                print("Could not convert curve due to arcs", curve)
                continue
            coordinates = curve.Points.CoordList
            points = []
            for i, segment in enumerate(curve.Segments):
                segment = segment.wrappedValue
                if i == 0:
                    points.append(self.file.createIfcCartesianPoint(coordinates[segment[0] - 1]))
                points.append(self.file.createIfcCartesianPoint(coordinates[segment[1] - 1]))
            polyline = self.file.create_entity("IfcPolyline", points)
            curve_map[curve] = polyline

        for curve, polyline in curve_map.items():
            for inverse in self.file.get_inverse(curve):
                ifcopenshell.util.element.replace_attribute(inverse, curve, polyline)



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021, 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.api
import ifcopenshell.guid
import ifcopenshell.util.selector
from typing import Union
from logging import Logger


class Patcher:
    def __init__(self, src: str, file: ifcopenshell.file, logger: Logger, query: str = "IfcWall"):
        """Extract certain elements into a new model

        Extract a subset of elements from an existing IFC data set and save it
        to a new IFC file. For example, you might want to extract only the walls
        in a model and save it as a new model.

        :param query: A query to select the subset of IFC elements.
        :type query: str

        Example:

        .. code:: python

            # Extract all walls
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ExtractElements", "arguments": ["IfcWall"]})

            # Extract all slabs
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ExtractElements", "arguments": ["IfcSlab"]})

            # Extract all walls and slabs
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ExtractElements", "arguments": ["IfcWall, IfcSlab"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.query = query

    def patch(self):
        self.contained_ins: dict[str, set[ifcopenshell.entity_instance]] = {}
        self.aggregates: dict[str, set[ifcopenshell.entity_instance]] = {}
        self.new = ifcopenshell.file(schema=self.file.wrapped_data.schema)
        self.owner_history = None
        self.reuse_identities: dict[int, ifcopenshell.entity_instance] = {}
        for owner_history in self.file.by_type("IfcOwnerHistory"):
            self.owner_history = self.new.add(owner_history)
            break
        self.add_element(self.file.by_type("IfcProject")[0])
        for element in ifcopenshell.util.selector.filter_elements(self.file, self.query):
            self.add_element(element)
        self.create_spatial_tree()
        self.file = self.new

    def add_element(self, element: ifcopenshell.entity_instance) -> None:
        new_element = self.append_asset(element)
        if not new_element:
            return
        self.add_spatial_structures(element, new_element)
        self.add_decomposition_parents(element, new_element)

    def append_asset(self, element: ifcopenshell.entity_instance) -> Union[ifcopenshell.entity_instance, None]:
        try:
            return self.new.by_guid(element.GlobalId)
        except:
            pass
        if element.is_a("IfcProject"):
            return self.new.add(element)
        return ifcopenshell.api.run(
            "project.append_asset", self.new, library=self.file, element=element, reuse_identities=self.reuse_identities
        )

    def add_spatial_structures(
        self, element: ifcopenshell.entity_instance, new_element: ifcopenshell.entity_instance
    ) -> None:
        """element is IfcElement"""
        for rel in getattr(element, "ContainedInStructure", []):
            spatial_element = rel.RelatingStructure
            new_spatial_element = self.append_asset(spatial_element)
            self.contained_ins.setdefault(spatial_element.GlobalId, set()).add(new_element)
            self.add_decomposition_parents(spatial_element, new_spatial_element)

    def add_decomposition_parents(
        self, element: ifcopenshell.entity_instance, new_element: ifcopenshell.entity_instance
    ) -> None:
        """element is IfcObjectDefinition"""
        for rel in element.Decomposes:
            parent = rel.RelatingObject
            new_parent = self.append_asset(parent)
            self.aggregates.setdefault(parent.GlobalId, set()).add(new_element)
            self.add_decomposition_parents(parent, new_parent)
            self.add_spatial_structures(parent, new_parent)

    def create_spatial_tree(self) -> None:
        for relating_structure_guid, related_elements in self.contained_ins.items():
            self.new.createIfcRelContainedInSpatialStructure(
                ifcopenshell.guid.new(),
                self.owner_history,
                None,
                None,
                list(related_elements),
                self.new.by_guid(relating_structure_guid),
            )
        for relating_object_guid, related_objects in self.aggregates.items():
            self.new.createIfcRelAggregates(
                ifcopenshell.guid.new(),
                self.owner_history,
                None,
                None,
                self.new.by_guid(relating_object_guid),
                list(related_objects),
            )



# IfcOpenShell - IFC toolkit and geometry engine
# Copyright (C) 2022 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcOpenShell.
#
# IfcOpenShell is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcOpenShell is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcOpenShell.  If not, see <http://www.gnu.org/licenses/>.

import os
import pytest
import ifcpatch
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util.element
import ifcopenshell.util.representation
import test.bootstrap


class TestMergeDuplicateTypes(test.bootstrap.IFC4):
    def test_run(self):
        context = self.file.createIfcGeometricRepresentationContext()

        wall1 = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        wall_type1 = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWallType", name="WallType")
        rep = self.file.createIfcShapeRepresentation(ContextOfItems=context)
        ifcopenshell.api.run("geometry.assign_representation", self.file, product=wall_type1, representation=rep)
        ifcopenshell.api.run("material.assign_material", self.file, products=[wall_type1], type="IfcMaterialLayerSet")
        ifcopenshell.api.run(
            "type.assign_type",
            self.file,
            related_objects=[wall1],
            relating_type=wall_type1,
            should_map_representations=False,
        )

        wall2 = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWall")
        wall_type2 = ifcopenshell.api.run("root.create_entity", self.file, ifc_class="IfcWallType", name="WallType")
        rep = self.file.createIfcShapeRepresentation(ContextOfItems=context)
        ifcopenshell.api.run("geometry.assign_representation", self.file, product=wall_type2, representation=rep)
        ifcopenshell.api.run("material.assign_material", self.file, products=[wall_type2], type="IfcMaterialLayerSet")
        ifcopenshell.api.run(
            "type.assign_type",
            self.file,
            related_objects=[wall2],
            relating_type=wall_type2,
            should_map_representations=False,
        )

        output = ifcpatch.execute({"file": self.file, "recipe": "MergeDuplicateTypes", "arguments": ["Name"]})

        assert len(output.by_type("IfcWall")) == 2
        wall_types = output.by_type("IfcWallType")
        assert len(wall_types) == 1
        wall_type = wall_types[0]
        assert set(ifcopenshell.util.element.get_types(wall_type)) == set(output.by_type("IfcWall"))

        # test not to remap representation
        assert ifcopenshell.util.representation.get_representation(wall1, context=context) == None
        assert ifcopenshell.util.representation.get_representation(wall2, context=context) == None

        # and not to create material usages
        assert ifcopenshell.util.element.get_material(wall1, should_inherit=False) == None
        assert ifcopenshell.util.element.get_material(wall2, should_inherit=False) == None


class TestMergeDuplicateTypesIFC2X3(test.bootstrap.IFC2X3, TestMergeDuplicateTypes):
    pass



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import ifcopenshell
import ifcopenshell.util.unit
import ifcopenshell.util.element
import ifcopenshell.util.geolocation
from ifcpatch.recipes.SetFalseOrigin import Patcher as SetFalseOrigin
from typing import Union
from logging import Logger


class Patcher:
    def __init__(self, src: str, file: ifcopenshell.file, logger: Logger, filepath: Union[str, ifcopenshell.file]):
        """Merge two IFC models into one

        Note that other than combining the two IfcProject elements into one, no
        further processing will be done. This means that you may end up with
        duplicate spatial hierarchies (i.e. 2 sites, 2 buildings, etc).

        Will automatically convert length units in the second model to the main
        model's unit before merging.

        :param filepath: The filepath of the second IFC model to merge into the
            first. The first model is already specified as the input to
            IfcPatch.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "MergeProject", "arguments": ["/path/to/model2.ifc"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.filepath = filepath

    def patch(self):
        if isinstance(self.filepath, ifcopenshell.file):
            other = self.filepath
        else:
            other = ifcopenshell.open(self.filepath)

        if (main_unit := self.get_unit_name(self.file)) != self.get_unit_name(other):
            other = ifcopenshell.util.unit.convert_file_length_units(other, main_unit)

        existing_origin = np.array(
            ifcopenshell.util.geolocation.auto_xyz2enh(self.file, 0, 0, 0, should_return_in_map_units=False)
        )
        other_origin = np.array(
            ifcopenshell.util.geolocation.auto_xyz2enh(other, 0, 0, 0, should_return_in_map_units=False)
        )

        existing_angle = ifcopenshell.util.geolocation.get_grid_north(self.file)
        other_angle = ifcopenshell.util.geolocation.get_grid_north(other)

        model_rotation = existing_angle - other_angle
        if model_rotation > 180:
            model_rotation = (360 - model_rotation) * -1
        elif model_rotation < -180:
            model_rotation = (model_rotation * -1) - 360

        if not np.allclose(existing_origin, other_origin) or not np.isclose(existing_angle, other_angle):
            x, y, z = ifcopenshell.util.geolocation.auto_enh2xyz(
                other, *existing_origin, is_specified_in_map_units=False
            )
            e, n, h = existing_origin
            SetFalseOrigin(
                "",
                other,
                self.logger,
                name="",
                x=x,
                y=y,
                z=z,
                e=e,
                n=n,
                h=h,
                gn_angle=existing_angle,
                rotate_angle=model_rotation,
            ).patch()

        self.existing_contexts: list[ifcopenshell.entity_instance] = self.file.by_type(
            "IfcGeometricRepresentationContext"
        )
        self.added_contexts: set[ifcopenshell.entity_instance] = set()

        original_project = self.file.by_type("IfcProject")[0]
        merged_project = self.file.add(other.by_type("IfcProject")[0])

        for element in other.by_type("IfcGeometricRepresentationContext"):
            new = self.file.add(element)
            self.added_contexts.add(new)

        for element in other:
            self.file.add(element)

        for inverse in self.file.get_inverse(merged_project):
            ifcopenshell.util.element.replace_attribute(inverse, merged_project, original_project)
        self.file.remove(merged_project)

        self.reuse_existing_contexts()

    def get_unit_name(self, ifc_file: ifcopenshell.file) -> str:
        length_unit = ifcopenshell.util.unit.get_project_unit(ifc_file, "LENGTHUNIT")
        return ifcopenshell.util.unit.get_full_unit_name(length_unit)

    def reuse_existing_contexts(self):
        to_delete = set()

        for added_context in self.added_contexts:
            equivalent_existing_context = self.get_equivalent_existing_context(added_context)
            if equivalent_existing_context:
                for inverse in self.file.get_inverse(added_context):
                    if self.file.schema != "IFC2X3":
                        if inverse.is_a("IfcCoordinateOperation"):
                            to_delete.add(inverse.id())
                            continue
                    ifcopenshell.util.element.replace_attribute(inverse, added_context, equivalent_existing_context)
                to_delete.add(added_context.id())

        for element_id in to_delete:
            try:
                ifcopenshell.util.element.remove_deep2(self.file, self.file.by_id(element_id))
            except:
                pass

    def get_equivalent_existing_context(
        self, added_context: ifcopenshell.entity_instance
    ) -> Union[ifcopenshell.entity_instance, None]:
        for context in self.existing_contexts:
            if context.is_a() != added_context.is_a():
                continue
            if context.is_a("IfcGeometricRepresentationSubContext"):
                if (
                    context.ContextType == added_context.ContextType
                    and context.ContextIdentifier == added_context.ContextIdentifier
                    and context.TargetView == added_context.TargetView
                ):
                    return context
            elif (
                context.ContextType == added_context.ContextType
                and context.ContextIdentifier == added_context.ContextIdentifier
            ):
                return context



# IfcPatch - IFC patching utiliy
# Copyright (C) 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.


class Patcher:
    def __init__(self, src, file, logger):
        """Allow ArchiCAD IFC spaces to open as Revit rooms

        The underlying problem is that Revit does not bring in IFC spaces as
        spaces / rooms in Revit when you link an IFC in Revit. This has been
        broken for at least 3 years and counting. This is a problem typically
        for ArchiCAD architects who want to send rooms to MEP folks using
        Revit.

        See bug: https://github.com/Autodesk/revit-ifc/issues/15

        The solution is to open an IFC in Revit instead of linking it, which
        will convert IFC spaces into Revit rooms. However, there are very
        specific scenarios where Revit will convert these rooms, which have
        been painstakingly reverse engineered through trial and error.
        Firstly, the rooms should have a lower bound with a Z value matching
        the Z value of the storey it is on. Secondly, although faceted breps
        do work in some scenarios (I assume Revit has an internal topological
        analysis tool), conversion to an extruded area solid yield much more
        robust results. Finally, changing the Precision value to an obscene
        number very strangely seems to cause a lot more rooms to be converted
        successfully.

        This patch is designed to only work on ArchiCAD IFC exports where the
        only contents of the IFC is IFC space and `nothing else`. It also
        requires you to run it using Blender, as the geometric modification
        uses the Blender geometry engine.

        Example:

        .. code:: python

            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "FixArchiCADToRevitSpaces", "arguments": []})
        """
        self.src = src
        self.file = file
        self.logger = logger

    def patch(self):
        import bpy
        import blenderbim.tool as tool
        import ifcopenshell
        import ifcopenshell.util.element
        from blenderbim.bim.ifc import IfcStore
        from mathutils import Vector, Matrix

        if len(bpy.data.objects) > 0:
            bpy.data.batch_remove(bpy.data.objects)
            bpy.ops.outliner.orphans_purge(do_local_ids=True, do_linked_ids=True, do_recursive=True)

        bpy.ops.bim.load_project(filepath=self.src)

        def recalculate_origin(wall):
            new_origin = wall.matrix_world @ Vector(wall.bound_box[0])
            if (wall.matrix_world.translation - new_origin).length < 0.001:
                return
            wall.data.transform(
                Matrix.Translation(
                    (wall.matrix_world.inverted().to_quaternion() @ (wall.matrix_world.translation - new_origin))
                )
            )
            wall.matrix_world.translation = new_origin

        for obj in bpy.context.visible_objects:
            bpy.context.view_layer.update()
            if "IfcSpace" not in obj.name:
                continue

            element = tool.Ifc.get_entity(obj)
            storey = ifcopenshell.util.element.get_aggregate(element)
            storey_obj = tool.Ifc.get_object(storey)
            target_z = storey_obj.location.z

            local_target_z = (obj.matrix_world.inverted() @ Vector((0, 0, target_z))).z
            local_target_zup = (obj.matrix_world.inverted() @ Vector((0, 0, target_z + 3))).z
            for v in obj.data.vertices:
                if round(v.co.z, 3) == 0:
                    v.co.z = local_target_z
            bpy.context.view_layer.update()
            recalculate_origin(obj)
            obj.select_set(True)

        bpy.ops.bim.update_representation(
            ifc_representation_class="IfcExtrudedAreaSolid/IfcArbitraryProfileDefWithVoids"
        )
        for context in IfcStore.get_file().by_type("IfcGeometricRepresentationContext", include_subtypes=False):
            if context.Precision:
                context.Precision = 10

        self.file = IfcStore.get_file()



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.api
import ifcopenshell.api.owner.settings
import ifcopenshell.util.pset
import ifcopenshell.util.element
import ifcopenshell.util.unit
from logging import Logger


class Patcher:
    def __init__(
        self,
        src: str,
        file: ifcopenshell.file,
        logger: Logger,
        unit: str = "METER",
    ):
        """Converts the length unit of a model to the specified unit

        Allowed metric units include METER, MILLIMETER, CENTIMETER, etc.
        Allowed imperial units include INCH, FOOT, MILE.

        :param unit: The name of the desired unit, defaults to "METER"
        :type unit: str

        Example:

        .. code:: python

            # Convert to millimeters
            model = ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ConvertLengthUnit", "arguments": ["MILLIMETER"]})

            # Convert to feet
            model = ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "ConvertLengthUnit", "arguments": ["FOOT"]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.unit = unit
        self.file_patched: ifcopenshell.file

    def patch(self):
        self.file_patched = ifcopenshell.util.unit.convert_file_length_units(self.file, self.unit)



#!/usr/bin/env python3

# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021, 2023 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import ifcopenshell
import logging
import typing
import inspect
import collections
import importlib
from typing import Union


__version__ = version = "0.7.10"


def execute(args: dict) -> Union[ifcopenshell.file, str]:
    """Execute a patch recipe

    The details of how the patch recipe is executed depends on the definition of
    the recipe, as well as the arguments passed to the recipe. See the
    documentation for each patch recipe separately to understand more.

    :param args: A dictionary of arguments, corresponding to the parameters
        listed subsequent to this in this docstring.
    :type args: dict
    :param input: A filepath to the incoming IFC file.
    :type input: str
    :param file: An IFC model to apply the patch recipe to.
    :type file: ifcopenshell.file.file
    :param recipe: The name of the recipe. This is the same as the filename of
        the recipe. E.g. "ExtractElements".
    :type recipe: str
    :param log: A filepath to a logfile.
    :type log: str,optional
    :param arguments: A list of zero or more positional arguments, depending on
        the patch recipe. Some patch recipes will require you to specify
        arguments, some won't.
    :type arguments: list
    :return: The result of the patch. This is typically a patched model, either
        as an object or as a string.
    :rtype: ifcopenshell.file.file,str

    Example:

    .. code:: python

        output = ifcpatch.execute({
            "input": "input.ifc",
            "file": ifcopenshell.open("input.ifc"),
            "recipe": "ExtractElements",
            "arguments": [".IfcWall"],
        })
        ifcpatch.write(output, "output.ifc")
    """
    if "log" in args:
        logging.basicConfig(filename=args["log"], filemode="a", level=logging.DEBUG)
    logger = logging.getLogger("IFCPatch")
    recipes = getattr(__import__("ifcpatch.recipes.{}".format(args["recipe"])), "recipes")
    recipe = getattr(recipes, args["recipe"])
    arguments = args.get("arguments", [])
    if recipe.Patcher.__init__.__doc__ is not None:
        patcher = recipe.Patcher(args.get("input"), args["file"], logger, *arguments)
    else:
        patcher = recipe.Patcher(args.get("input"), args["file"], logger, arguments)
    patcher.patch()
    output = getattr(patcher, "file_patched", patcher.file)
    return output


def write(output: Union[ifcopenshell.file, str], filepath: str) -> None:
    """Write the output of an IFC patch to a file

    Typically a patch output would be a patched IFC model file object, or as a
    string. This function lets you agnostically write that output to a filepath.

    :param output: The results from ifcpatch.execute()
    :type output: ifcopenshell.file.file,str
    :param filepath: A filepath to where the results of the patched model should
        be written to.
    :type filepath: str
    :return: None
    :rtype: None
    """
    if output is None:
        return
    elif isinstance(output, str):
        if os.path.exists(output):
            shutil.move(output, filepath)
        else:
            with open(filepath, "w") as text_file:
                text_file.write(output)
    else:
        output.write(filepath)


def extract_docs(
    submodule_name: str, cls_name: str, method_name: str = "__init__", boilerplate_args: typing.Iterable[str] = None
):
    """Extract class docstrings and method arguments

    :param submodule_name: Submodule from which to extract the class
    :param cls_name: Class from which to extract the docstring and method arguments
    :param method_name: Class Method name from which to extract arguments
    :param boilerplate_args: String iterable containing arguments that shall not be parsed
    """
    spec = importlib.util.spec_from_file_location(
        submodule_name, f"{os.path.dirname(inspect.getabsfile(inspect.currentframe()))}/recipes/{submodule_name}.py"
    )
    submodule = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(submodule)
        try:
            return _extract_docs(getattr(submodule, cls_name), method_name, boilerplate_args)
        except AttributeError as e:
            print(e)
    except ModuleNotFoundError as e:
        print(f"Error : IFCPatch {str(submodule)} could not load because : {str(e)}")


def _extract_docs(cls, method_name, boilerplate_args):
    inputs = collections.OrderedDict()
    method = getattr(cls, method_name)
    docs = {"class": cls}
    if boilerplate_args is None:
        boilerplate_args = []

    signature = inspect.signature(method)
    for name, parameter in signature.parameters.items():
        if name == "self" or name in boilerplate_args:
            continue
        inputs[name] = {"name": name}
        if isinstance(parameter.default, (str, float, int, bool)):
            inputs[name]["default"] = parameter.default

    type_hints = typing.get_type_hints(method)
    for input_name in inputs.keys():
        type_hint = type_hints.get(input_name, None)
        if type_hint is None:  # The argument is not type-hinted. (Or hinted to None ??)
            continue
        if isinstance(type_hint, typing._UnionGenericAlias):
            inputs[input_name]["type"] = [t.__name__ for t in typing.get_args(type_hint)]
        else:
            inputs[input_name]["type"] = type_hint.__name__

    description = ""
    doc = method.__doc__
    if doc is not None:
        for i, line in enumerate(doc.split("\n")):
            line = line.strip()
            if i == 0:
                docs["name"] = line
            elif line.startswith(":return:"):
                docs["output"] = {"name": line.split(":")[2].strip(), "description": line.split(":")[3].strip()}
            elif line.startswith(":param"):
                param_name = line.split(":")[1].strip().replace("param ", "")
                if param_name in inputs:
                    inputs[param_name]["description"] = line.split(":")[2].strip()
            elif i == 2:
                description += line
            elif i > 2:
                description += "\n" + line

        docs["description"] = description.strip()
    docs["inputs"] = inputs
    return docs



# IfcPatch - IFC patching utiliy
# Copyright (C) 2020, 2021 Dion Moult <dion@thinkmoult.com>
#
# This file is part of IfcPatch.
#
# IfcPatch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IfcPatch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IfcPatch.  If not, see <http://www.gnu.org/licenses/>.

import ifcopenshell
import ifcopenshell.guid
from logging import Logger


class Patcher:
    def __init__(self, src: str, file: ifcopenshell.file, logger: Logger, only_duplicates=False):
        """Regenerate GlobalIds in an IFC model

        All root elements in an IFC model must be identified by a unique Global
        ID (also known as a GUID or UUID). Some proprietary BIM software do this
        incorrect (I know right), either by generating an invalid ID, creating
        duplicate IDs, generating IDs in a way that is not universally unique or
        as random as you might prefer (e.g. non-compliant with UUID v4).

        This will regenerate new GlobalIds for the entire model.

        :param only_duplicates: If set to True, new GlobalIds will only be
            generated for duplicate IDs. This is a safe thing to run to ensure
            IFCs are valid. If False, all GlobalIds will be regenerated.
        :type only_duplicates: bool

        Example:

        .. code:: python

            # Regenerate all GlobalIds
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "RegenerateGlobalIds", "arguments": []})

            # Regenerate only duplicate GlobalIds
            ifcpatch.execute({"input": "input.ifc", "file": model, "recipe": "RegenerateGlobalIds", "arguments": [True]})
        """
        self.src = src
        self.file = file
        self.logger = logger
        self.only_duplicates = only_duplicates

    def patch(self):
        if self.only_duplicates:
            duplicates = 0
            invalid_ids = 0

            guids = set()
            for element in self.file.by_type("IfcRoot"):
                if element.GlobalId in guids:
                    element.GlobalId = ifcopenshell.guid.new()
                    duplicates += 1
                elif len(element.GlobalId) != 22 or element.GlobalId[0] not in "0123":
                    element.GlobalId = ifcopenshell.guid.new()
                    invalid_ids += 1
                else:
                    try:
                        ifcopenshell.guid.expand(element.GlobalId)
                    except:
                        element.GlobalId = ifcopenshell.guid.new()
                    invalid_ids += 1
                guids.add(element.GlobalId)

            print("Replaced %s duplicate GlobalIds" % duplicates)
            print("Replaced %s invalid GlobalIds" % invalid_ids)
        else:
            for element in self.file.by_type("IfcRoot"):
                element.GlobalId = ifcopenshell.guid.new()
